#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "webacl.h"
#include "charmap.h"


typedef struct {
    int     code;
    char    *message;
} error_t;

static const error_t errors[] = {
    {-10, "null argument not allowed"},
    {-20, "invalid entry prefix"},
    {-21, "invalid entity char"},
    {-22, "invalid permission char"},
    {-23, "empty entity name"},
    {0, NULL},
};

const char *
webacl_error(int code)
{
    int i;
    for(i=0; errors[i].code != 0; i++) {
        if(errors[i].code == code)
            return errors[i].message;
    }
    return "unknown error";
}


int
webacl_iter_start(webacl_t *acl, webacl_iter_t *iter)
{
    iter->acl = acl;
    iter->pos = acl;
    return 1;
}


int
webacl_iter_next(webacl_iter_t *iter, webacl_item_t *item)
{
    char *s;
    int i, has_name;

    if(iter == NULL || item == NULL)
        return -10;

    s = iter->pos;

    /* Consume whitespace at the entry start */
    for(; charmap[(int)*s] & (DELIM); s++);

    has_name = 0;
    switch(*s) {
        case 0:
            return 0;
        case WEBACL_USER:
            has_name = 1;
            item->kind = WEBACL_USER;
            break;
        case WEBACL_OWNER:
            item->kind = WEBACL_OWNER;
            break;
        case WEBACL_GROUP:
            has_name = 1;
            item->kind = WEBACL_GROUP;
            break;
        case WEBACL_AUTH:
            item->kind = WEBACL_AUTH;
            break;
        case WEBACL_ANON:
            item->kind = WEBACL_ANON;
            break;
        default:
            return -20;
    }

    /* Consume colon between kind and name.
     */
    s++;
    if(*s != ':') {
        return -20;
    }
    s++;
    
    /* Store entity name and length to ent/ent_l */
    if(has_name) {
        item->ent = s;
        for(i=0; charmap[(int)*s] & ENTITY; i++, s++);
        item->ent_l = i;
        if(i < 1) {
            return -23;
        }
        if(*s != ':') {
            return -21;
        }
        s++;
    } else {
        item->ent_l = 0;
    }

    /* Store permission string and length to perm/perm_l */
    item->perm = s;
    for(i=0; charmap[(int)*s] & PERM; i++, s++);
    item->perm_l = i;
    if(charmap[(int)*s] & (DELIM|END)) {
        iter->pos = s;
    } else {
        return -22;
    }
    return 1;
}

void
webacl_iter_stop(webacl_iter_t *iter)
{
}



int
webacl_parse(webacl_t *acl, int *len)
{
    webacl_iter_t iter;
    webacl_item_t item;
    int r, l=0;

    if(acl == NULL)
        return -10;
    webacl_iter_start(acl, &iter);
    while(1) {
        r = webacl_iter_next(&iter, &item);
        if (r != 1) {
            webacl_iter_stop(&iter);
            return r;
        }
        l++;
    }
    webacl_iter_stop(&iter);
    if(len != NULL) {
        *len = l;
    }
    return 0;
}


int
webacl_check(webacl_t *acl, char *user, char **groups, char perm)
{
    return webacl_check_ex(acl, user, groups, NULL, perm);
}


int
webacl_check_ex(
    webacl_t *acl, char *user, char **groups, char *owner, char perm
)
{
    webacl_iter_t iter;
    webacl_item_t item;

    char **g, *s;
    int r, i, found;

    if(acl == NULL)
        return -10;
    webacl_iter_start(acl, &iter);
    while(1) {
        r = webacl_iter_next(&iter, &item);
        if (r == 0) {
            /* end */
            webacl_iter_stop(&iter);
            break;
        } else if (r != 1) {
            /* syntax error */
            webacl_iter_stop(&iter);
            return r;
        }
        switch (item.kind) {
            case WEBACL_USER:
                if(user == NULL)
                    continue;
                if(strncmp(user, item.ent, item.ent_l) != 0)
                    continue;
                if(user[item.ent_l] != 0)
                    continue;
                break;
            case WEBACL_OWNER:
                if(user == NULL)
                    continue;
                if(owner == NULL)
                    continue;
                if(strcmp(user, owner) != 0)
                    continue;
                break;
            case WEBACL_GROUP:
                if(user == NULL)
                    continue;
                if(groups == NULL)
                    continue;
                found = 0;
                for(g=groups; *g != 0; ++g) {
                    if(strncmp(*g, item.ent, item.ent_l) != 0)
                        continue;
                    if((*g)[item.ent_l] != 0)
                        continue;
                    found = 1;
                    break;
                }
                if(!found)
                    continue;
                break;
            case WEBACL_AUTH:
                if(user == NULL)
                    continue;
                break;
            case WEBACL_ANON:
                break;
        }
        for(i=0, s=item.perm; i<item.perm_l; s++, i++) {
            if((*s & 0x5f) == perm) {
                webacl_iter_stop(&iter);
                if(*s == perm) {
                    return (int)item.kind;
                } else {    
                    return -((int)item.kind);
                }
            }

        }
    }
    webacl_iter_stop(&iter);
    return 0;
}


int
webacl_escape(const char *src, char *dest, size_t dest_l)
{
    int i, j, l;
    char c;

    if(src == NULL || dest == NULL)
        return -1;
    l = strlen(src);
    for(i=0, j=0; i<l && j<dest_l; i++, j++) {
        c = src[i];
        if(c == '%') {
            dest[j++] = '%';
            dest[j++] = '2';
            dest[j] = '5';
        } else if(c & ENTITY) {
            dest[j] = c;
        } else {
            /* XXX */
            dest[j++] = '%';
            dest[j++] = '0' + (c / 16);
            dest[j] = '0' + (c % 16);
        }
    }
    return 0;
}

int
webacl_unescape(const char *src, char *dest, size_t dest_l)
{
    return 0;
}
