#include "Python.h"
#include "webacl.h"


typedef struct {
    PyObject_HEAD
    webacl_t  *acl;
} webacl;

static PyTypeObject webacltype;


static webacl *
webacl_new(PyTypeObject *type, PyObject *args)
{
    webacl_t *s;
    webacl   *acl;
    int         r;

    if(! PyArg_ParseTuple(args, "s", &s)) {
        return NULL;
    }
    r = webacl_parse(s, NULL);
    if(r != 0) {
        PyErr_SetString(PyExc_ValueError, webacl_error(r));
        return NULL;
    }
    if(!(acl = PyObject_NEW(webacl, &webacltype))) {
        return NULL;
    }
    acl->acl = strdup(s);
    return acl;
}

static void
webacl_dealloc(webacl *self)
{
    if(self->acl) free(self->acl);
}


static PyObject *
py_webacl_items(webacl *self, PyObject *args)
{
    PyObject        *items, *tup;
    webacl_iter_t   iter;
    webacl_item_t   item;

    if (!PyArg_ParseTuple(args, ":items"))
        return NULL;

    items = PyList_New(0);
    webacl_iter_start(self->acl, &iter);
    while(1) {
        if (webacl_iter_next(&iter, &item) != 1)
            break;
        tup = PyTuple_New(3);
        PyTuple_SET_ITEM(tup, 0, PyString_FromStringAndSize(&item.kind, 1));
        if(item.ent_l > 0) {
            PyTuple_SET_ITEM(tup, 1,
                             PyString_FromStringAndSize(item.ent, item.ent_l));
        } else {
            Py_INCREF(Py_None);
            PyTuple_SET_ITEM(tup, 1, Py_None);
        }
        PyTuple_SET_ITEM(tup, 2,
                         PyString_FromStringAndSize(item.perm, item.perm_l));
         PyList_Append(items, tup);
    }
    webacl_iter_stop(&iter);
    return items;
}


static PyObject *
py_webacl_check(webacl *self, PyObject *args, PyObject *kw)
{
    PyObject *_groups=NULL, *obj;
    char *user = NULL;
    char **groups = NULL;
    char *perm_string = NULL, perm;
    int perm_length, group_num = 0, i, res;
    char *owner = NULL;

    static char *kwlist[] = {"user", "groups", "perm", "owner", NULL};

    if (!PyArg_ParseTupleAndKeywords(
                args, kw, "zOs#|z", kwlist, 
                &user, &_groups, &perm_string, &perm_length, &owner))
        return NULL;
    if(perm_length != 1) {
        PyErr_SetString(PyExc_ValueError, "invalid perm length");
        goto onerror;
    }
    perm = perm_string[0];

    if(_groups) {
        if(PyObject_IsInstance(_groups, (PyObject*)&PyList_Type)) {
            group_num = PySequence_Length(_groups);
        } else if(PyObject_IsInstance(_groups, (PyObject*)&PyTuple_Type)) {
            group_num = PySequence_Length(_groups);
        } else {
            PyErr_SetString(PyExc_ValueError, "list of strings required");
            goto onerror;
        }
        for(i=0; i<group_num; i++) {
            obj = PySequence_Fast_GET_ITEM(_groups,i);
            if(!PyObject_IsInstance(obj, (PyObject*)&PyString_Type)) {
                PyErr_SetString(PyExc_ValueError, "invalid group type");
                goto onerror;
            }
        }
        if(group_num) {
            groups = malloc(sizeof(char * ) * (group_num+1));
            for(i=0; i<group_num; i++) {
                obj = PySequence_Fast_GET_ITEM(_groups,i);
                groups[i] = PyString_AsString(obj);
            }
            groups[group_num] = NULL;
        }
    }
    res = webacl_check_ex(self->acl, user, groups, owner, perm);
    /* 
    printf("acl:<%s>\n", self->acl);
    printf("user:<%s>\n", user);
    printf("perm:<%c>\n", perm);
    printf("owner:<%s>\n", owner);
    printf("group_num:<%d>\n", group_num);
    for(i=0; i<group_num; i++) {
        printf("group[%d]:<%s>\n", i, groups[i]);
    }
    printf("result:<%d>\n", res);
    */
    if(groups)
        free(groups);
    if(res > 0) {
        Py_INCREF(Py_True);
        return Py_True;
    } else {
        Py_INCREF(Py_False);
        return Py_False;
    }

onerror:
    if(groups)
        free(groups);
    return NULL;
}


static int
py_webacl_comp(webacl *left, webacl *right)
{
    return strcmp(left->acl, right->acl);
}


static long
py_webacl_hash(webacl *ob)
{
    register Py_ssize_t len;
    register unsigned char *p;
    register long x;

    len = strlen(ob->acl);
    p = (unsigned char *) ob->acl;
    x = *p << 7;
    while (--len >= 0)
        x = (1000003*x) ^ *p++;
    x ^= strlen(ob->acl);
    if (x == -1)
        x = -2;
    return x;
}



static PyObject *
py_webacl__repr__(webacl *self)
{
    PyObject *ret = PyString_FromFormat("webacl('%s')", self->acl);
    return ret;
}


static PyObject *
py_webacl__str__(webacl *self)
{
    PyObject *ret = PyString_FromString(self->acl);
    return ret;
}


static PyMethodDef webacl_methods[] = {
    {"items",  (PyCFunction) py_webacl_items, METH_VARARGS,
                PyDoc_STR("webacl.items()")},
    {"check",  (PyCFunction) py_webacl_check, METH_VARARGS | METH_KEYWORDS,
                PyDoc_STR("webacl.check(user, groups, perm, owner=None)")},
    {NULL,	NULL},
};


static PyTypeObject webacltype = {
	PyObject_HEAD_INIT(&PyType_Type)
	0,
	"webacl.webacl",
	sizeof(webacl),
	0,
	(destructor) webacl_dealloc,		/* tp_dealloc */
	0,					/* tp_print */
	0,					/* tp_getattr */
	0,					/* tp_setattr */
	(cmpfunc)py_webacl_comp,		/* tp_compare */
	(reprfunc)py_webacl__repr__,		/* tp_repr */
	0,					/* tp_as_number */
	0,					/* tp_as_sequence */
	0,					/* tp_as_mapping */
	(hashfunc)py_webacl_hash,		/* tp_hash */
	0,					/* tp_call */
	(reprfunc)py_webacl__str__,		/* tp_str */
	0,					/* tp_getattro */
	0,					/* tp_setattro */
	0,					/* tp_as_buffer */
	Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE, /* tp_flags */
	0,					/* tp_doc */
	0,					/* tp_traverse */
	0,					/* tp_clear */
	0,	                                /* tp_richcompare */
	0,					/* tp_weaklistoffset */
	0,					/* tp_iter */
	0,					/* tp_iternext */
	webacl_methods,			        /* tp_methods */
	0,					/* tp_members */
	0,			                /* tp_getset */
	0,                                   	/* tp_base */
	0,					/* tp_dict */
	0,					/* tp_descr_get */
	0,					/* tp_descr_set */
	0,					/* tp_dictoffset */
        0,		                        /* tp_init */
	0,					/* tp_alloc */
	(newfunc) webacl_new,	                /* tp_new */
};


PyMODINIT_FUNC
initwebacl(void)
{
    PyObject *m;

    m = Py_InitModule("webacl", NULL);
    if (m == NULL)
        return;

    if (PyType_Ready(&webacltype) < 0)
        return;

    Py_INCREF(&webacltype);
    if (PyModule_AddObject(m, "webacl", (PyObject *) &webacltype) < 0)
        return;

    if (PyModule_AddObject(m, "WEBACL_USER", PyString_FromString("u")) < 0)
        return;
    if (PyModule_AddObject(m, "WEBACL_OWNER", PyString_FromString("o")) < 0)
        return;
    if (PyModule_AddObject(m, "WEBACL_GROUP", PyString_FromString("g")) < 0)
        return;
    if (PyModule_AddObject(m, "WEBACL_AUTH", PyString_FromString("a")) < 0)
        return;
    if (PyModule_AddObject(m, "WEBACL_ANON", PyString_FromString("*")) < 0)
        return;

}
